#!/usr/bin/env python3
""" coding: utf-8
# Il y a six types de fichiers à traiter.
#
# nb images | tz      | poni
# ----------|---------|
# 5         |  -1     | x
# 5         |  0      | scan3.poni
# 5         |  x(-1)  | x
# 5         |  x(0)   | scan3.poni
# 1         |  -1     | x
# 1         |  0      | scan3.poni

# In[7]:

# # extraction des scans avec 6 positions en tx, quelque soit tz.
# # LONG
# import glob

# files = glob.glob(os.path.join(ROOT, "*.nxs"))

# def is_ok(filename: str) -> bool:
#     with h5py.File(filename, mode='r') as f:
#         for imgs, tx, tz in zip(get_images(f), get_tx(f), get_tz(f)):
#             return True if tx.shape[0] == 6 else False

# good = [f for f in files if is_ok(f)]
# print(good)
"""

from typing import Iterator, List, NamedTuple, Text, Tuple, Union

import os

from functools import partial
from itertools import chain

import h5py
import numpy
import pylab
import pyFAI

from fabio.edfimage import edfimage
from numpy import ndarray

from pyFAI.goniometer import GeometryTransformation, GoniometerRefinement
from pyFAI.gui import jupyter

from common import *


ROOT = "/home/experiences/instrumentation/picca/jupyter/mars/20160800/"
PUBLISHED = os.path.join(ROOT, "published-data")

CALIB = os.path.join(ROOT, "scan_3_01.nxs")

# H5Path data constructors
MetaDataSource = NamedTuple("MetaDataSource", [("images", H5Path),
                                               ("tx", H5Path),
                                               ("tz", H5Path)])

MetaData = NamedTuple("MetaData", [("image", ndarray),
                                   ("tx", float),
                                   ("tz", float)])

_MultiCalibMarsTxTz = NamedTuple("_MultiCalibMarsTxTz",
                                 [("filename", Text),
                                  ("metasources", MetaDataSource),
                                  ("idxs", List[int]),
                                  ("calibrant", Text),
                                  ("detector", Text),
                                  ("wavelength", float)])

class MultiCalibMarsTxTz(_MultiCalibMarsTxTz):
    def __len__(self) -> int:
        with h5py.File(self.filename, mode='r') as f:
            return get_shape(f, self.metasources.images)[0]

    def __item(self, f: h5py.File, index: int) -> MetaData:
        return MetaData(get_item_at_index(f,
                                          self.metasources.images, index),
                        get_item_at_index(f,
                                          self.metasources.tx, index),
                        get_item_at_index(f,
                                          self.metasources.tz, index))

    def __item__(self, index: int) -> MetaData:
        with h5py.File(self.filename, mode='r') as f:
            return self.__item(f, index)

    def frames(self) -> Iterator[MetaData]:
        with h5py.File(self.filename, mode='r') as f:
            for index in self.idxs:
                yield self.__item(f, index)

    def all_frames(self) -> Iterator[MetaData]:
        with h5py.File(self.filename, mode='r') as f:
            for index in range(len(self)):
                yield self.__item(f, index)


def save_as_edf(calib: MultiCalibMarsTxTz,
                basedir: Text) -> None:
    """Save the multi calib images into edf files in order to do the first
    calibration"""
    for idx, metadata in zip(calib.idxs, calib.frames()):
        base = os.path.splitext(os.path.basename(calib.filename))[0]
        output = os.path.join(basedir, base + '_%d.edf' % (idx,))
        edfimage(metadata.image).write(output)


def optimize_with_new_images(multicalib: MultiCalibMarsTxTz,
                             gonioref: GoniometerRefinement,
                             calibrant: pyFAI.calibrant.Calibrant,
                             pts_per_deg: float=1) -> None:
    """This function adds new images to the pool of data used for the
    refinement.  A set of new control points are extractred and a
    refinement step is performed at each iteration The last image of
    the serie is displayed

    """
    sg = None
    for idx, metadata in enumerate(multicalib.all_frames()):
        print()
        base = os.path.splitext(os.path.basename(multicalib.filename))[0]

        label = base + "_%d" % (idx,)
        if label in gonioref.single_geometries:
            continue
        print(label)

        sg = gonioref.new_geometry(label, image=metadata.image,
                                   metadata=metadata, calibrant=calibrant)
        print(sg.extract_cp(pts_per_deg=pts_per_deg))
    print("*"*50)
    gonioref.refine2()
    if sg:
        sg.geometry_refinement.set_param(gonioref.get_ai(sg.get_position()).param)  # noqa
        jupyter.display(sg=sg)


def get_wavelength(multicalib: MultiCalibMarsTxTz) -> float:
    """Return the wavelength"""
    return multicalib.wavelength


def get_calibrant(multicalib: MultiCalibMarsTxTz) -> pyFAI.calibrant.Calibrant:
    """Return the calibrant with the right wavelength"""
    calibrant = pyFAI.calibrant.get_calibrant(multicalib.calibrant)
    calibrant.wavelength = get_wavelength(multicalib)
    return calibrant


def get_detector(multicalib: MultiCalibMarsTxTz) -> pyFAI.Detector:
    return pyFAI.detector_factory(multicalib.detector)


def calibration(json: str) -> None:
    """Do a calibration with a bunch of images"""

    wavelength = 4.85945727522e-11

    multicalib = MultiCalibMarsTxTz(os.path.join(ROOT, "scan_3_01.nxs"),
                                    MetaDataSource(H5PathWithAttribute("interpretation", b"image"),  # noqa
                                                   H5PathContains("scan_data/actuator_1_1"),  # noqa
                                                   H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", 0.0)),  # noqa
                                    [2, 5, 8], "LaB6", "xpad_flat", wavelength)

    multicalib2 = MultiCalibMarsTxTz(os.path.join(ROOT, "scan_4_01.nxs"),
                                     MetaDataSource(H5PathWithAttribute("interpretation", b"image"),  # noqa
                                                    H5PathContains("scan_data/actuator_1_1"),  # noqa
                                                    H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", -1.0)),  # noqa
                                     [], "LaB6", "xpad_flat", wavelength)

    # save all the ref as images in order to do the calibration with
    # pyFAI-calib[2].
    save_as_edf(multicalib, PUBLISHED)

    # Definition of the geometry refinement: the parameter order is
    # the same as the param_names
    calibrant = get_calibrant(multicalib)
    detector = get_detector(multicalib)

    distance = 0.258705917299
    poni1_scale = 0.001
    poni1_offset = 0.132825374721
    poni2_scale = 0.0012272727272727272
    poni2_offset = -0.9488181818181818
    rot1 = 0.00388272369359
    rot2 = -0.00942588451226
    rot3 = 7.19961198098e-07

    parameters = [Parameter("dist", distance, (distance, distance)),
                  Parameter("poni1_offset", poni1_offset, (0, 0.2)),
                  Parameter("poni1_scale", poni1_scale, (0, 0.002)),
                  Parameter("poni2_offset", poni2_offset, (-1, -0.7)),
                  Parameter("poni2_scale", poni2_scale, (-1, 1)),
                  Parameter("rot1", rot1, (rot1, rot1)),
                  Parameter("rot2", rot2, (rot2, rot2)),
                  Parameter("rot3", rot3, (rot3, rot3))]

    params = {p.name: p.value for p in parameters}
    bounds = {p.name: p.bounds for p in parameters}
    param_names = [p.name for p in parameters]

    # Let's refine poni1 and poni2 also as function of the distance:

    trans_function = GeometryTransformation(param_names=param_names,
                                            pos_names=["tx", "tz"],
                                            dist_expr="dist",
                                            poni1_expr="tz * poni1_scale + poni1_offset",  # noqa
                                            poni2_expr="tx * poni2_scale + poni2_offset",  # noqa
                                            rot1_expr="rot1",
                                            rot2_expr="rot2",
                                            rot3_expr="rot3")

    def pos_function(metadata: MetaData) -> Tuple[float, float]:
        """Definition of the function reading the detector position from the
        header of the image."""
        return metadata.tx, metadata.tz

    gonioref = GoniometerRefinement(params,  # initial guess
                                    bounds=bounds,
                                    pos_function=pos_function,
                                    trans_function=trans_function,
                                    detector=detector,
                                    wavelength=wavelength)

    print("Empty refinement object:")
    print(gonioref)

    # Let's populate the goniometer refinement object with the known poni
    for idx, metadata in zip(multicalib.idxs, multicalib.frames()):
        base = os.path.splitext(os.path.basename(multicalib.filename))[0]

        label = base + "_%d" % (idx,)
        control_points = os.path.join(PUBLISHED, base + "_%d.npt" % (idx,))
        ai = pyFAI.load(os.path.join(PUBLISHED, base + "_%d.poni" % (idx,)))  # noqa
        print(ai)

        gonioref.new_geometry(label, metadata.image, metadata,
                              control_points, calibrant, ai)

    print("Filled refinement object:")
    print(gonioref)
    print(os.linesep + "\tlabel \t tx")
    for k, v in gonioref.single_geometries.items():
        print(k, v.get_position())

    for g in gonioref.single_geometries.values():
        ai = gonioref.get_ai(g.get_position())
        print(ai)

    for sg in gonioref.single_geometries.values():
        jupyter.display(sg=sg)

    gonioref.refine2()

    for multi in [multicalib, multicalib2]:
        optimize_with_new_images(multi, gonioref, calibrant)

    # for idx, sg in enumerate(gonioref.single_geometries.values()):
    #    sg.geometry_refinement.set_param(gonioref.get_ai(sg.get_position()).param)
    #    jupyter.display(sg=sg)

    gonioref.save(json)

    # pylab.show()


def _integrate(json: Text, multicalib: Tuple[MultiCalibMarsTxTz, MultiCalibMarsTxTz]) -> None:
    # do not do the computation if the .dat already exist
    output = multicalib[0].filename + '.dat'
    if os.path.exists(output):
        return

    THRESHOLD = 12000
    """Integrate a file with a json calibration file"""
    gonio = pyFAI.goniometer.Goniometer.sload(json)

    images = []
    positions = []
    for metadata in chain(multicalib[0].all_frames(),
                          multicalib[1].all_frames()):
        images.append(metadata.image)
        positions.append((metadata.tx, metadata.tz))
    mai = gonio.get_mg(positions)

    # compute the mask
    detector = get_detector(multicalib[0])
    mask = numpy.array(detector.mask)
    lst_mask = []
    for img in images:  # remove all pixels above the threshold"
        if THRESHOLD is not None:
            mask_t = numpy.where(img > THRESHOLD, True, False)
            lst_mask.append(numpy.logical_or(mask, mask_t))
        else:
            lst_mask.append(mask)

    res = mai.integrate1d(images, 10000, lst_mask=lst_mask)
    numpy.savetxt(output, numpy.array(res).T)

    #jupyter.plot1d(res)
    #pylab.show()


def integrate(json: Text, mcals: List[Tuple[MultiCalibMarsTxTz, MultiCalibMarsTxTz]]) -> None:
    for mcal in mcals:
        try:
            print(mcal[0].filename, mcal[1].filename)
            _integrate(json, mcal)
        except:
            pass

def main():
    wavelength = 4.85945727522e-11
    lab6 = MultiCalibMarsTxTz(os.path.join(ROOT, "scan_3_01.nxs"),
                              MetaDataSource(H5PathWithAttribute("interpretation", b"image"),  # noqa
                                             H5PathContains("scan_data/actuator_1_1"),  # noqa
                                             H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", 0.0)),  # noqa
                              [2, 5, 8], "LaB6", "xpad_flat", wavelength)

    lab6_2 = MultiCalibMarsTxTz(os.path.join(ROOT, "scan_4_01.nxs"),
                                MetaDataSource(H5PathWithAttribute("interpretation", b"image"),  # noqa
                                               H5PathContains("scan_data/actuator_1_1"),  # noqa
                                               H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", -1.0)),  # noqa
                                [], "LaB6", "xpad_flat", wavelength)

    JSON = os.path.join(PUBLISHED, "calibration.json")
    #calibration(JSON, [lab6, lab6_2])

    # integration des echantillons
    tz1 = [ MultiCalibMarsTxTz(os.path.join(ROOT, "scan_%d_01.nxs" % (i,)),
                               MetaDataSource(H5PathWithAttribute("interpretation", b"image"),  # noqa"
                                              H5PathContains("scan_data/actuator_1_1"),  # noqa
                                              H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", -1.0)),  # noqa
                               [], "LaB6", "xpad_flat", wavelength)
            for i in [77, 79, 81, 83, 85, 87, 89, 91]]


    tz0 = [ MultiCalibMarsTxTz(os.path.join(ROOT, "scan_%d_01.nxs" % (i,)),
                               MetaDataSource(H5PathWithAttribute("interpretation", b"image"),  # noqa"
                                              H5PathContains("scan_data/actuator_1_1"),  # noqa
                                              H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", 0.0)),  # noqa
                               [], "LaB6", "xpad_flat", wavelength)
            for i in [78, 80, 82, 84, 86, 88, 90, 92]]

    tz3 = [ MultiCalibMarsTxTz(os.path.join(ROOT, "scan_%d_01.nxs" % (i,)),
                               MetaDataSource(H5PathWithAttribute("interpretation", b"image"),  # noqa"
                                              H5PathContains("scan_data/actuator_1_1"),  # noqa
                                              H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", 0.0)),  # noqa
                               [], "LaB6", "xpad_flat", wavelength)
            for i in range(399, 588, 2) if i not in [523, 527, 581]]

    tz5 = [ MultiCalibMarsTxTz(os.path.join(ROOT, "scan_%d_01.nxs" % (i,)),
                               MetaDataSource(H5PathWithAttribute("interpretation", b"image"),  # noqa"
                                              H5PathContains("scan_data/actuator_1_1"),  # noqa
                                              H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", -5.0)),  # noqa
                               [], "LaB6", "xpad_flat", wavelength)
            for i in range(400, 589, 2) if i not in [524, 528, 582]]

    # samples = [(lab6, lab6_2)]  + list(zip(tz1, tz0)) + list(zip(tz3, tz5))
    samples = list(zip(tz3, tz5))
    integrate(JSON, samples)
    pylab.show()

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
""" coding: utf-8
"""

from typing import Iterator, List, NamedTuple, Text, Tuple, Union

import os

from functools import partial
from math import pi, radians

import h5py
import pylab
import pyFAI

from fabio.edfimage import edfimage
from numpy import ndarray

from pyFAI.goniometer import GeometryTransformation, GoniometerRefinement
from pyFAI.gui import jupyter

from common import *

ROOT = "/home/experiences/diffabs/reguer/20170063/"
PUBLISHED = os.path.join(ROOT, "published-data")
CALIB = os.path.join(PUBLISHED, "xrd", "calibrationCeO2")

MetaDataSource = NamedTuple("MetaDataSource", [("image", H5Path),
                                               ("delta", H5Path)])

MetaData = NamedTuple("MetaData", [("image", ndarray),
                                   ("delta", float)])

_Diffabs = NamedTuple("_Diffabs", [("filename", Text),
                                   ("basedir", Text),
                                   ("metasources", MetaDataSource),
                                   ("idxs", List[int]),
                                   ("calibrant", Text),
                                   ("detector", Text),
                                   ("wavelength", float)])


class Diffabs(_Diffabs):
    def __len__(self) -> int:
        with h5py.File(self.filename, mode='r') as f:
            return get_shape(f, self.metasources.image)[0]

    def __item(self, f: h5py.File, index: int) -> MetaData:
        return MetaData(get_item_at_index(f,
                                          self.metasources.image, index),
                        get_item_at_index(f,
                                          self.metasources.delta, index))

    def __item__(self, index: int) -> MetaData:
        with h5py.File(self.filename, mode='r') as f:
            return self.__item(f, index)

    def frames(self) -> Iterator[MetaData]:
        with h5py.File(self.filename, mode='r') as f:
            for index in self.idxs:
                yield self.__item(f, index)

    def all_frames(self) -> Iterator[MetaData]:
        with h5py.File(self.filename, mode='r') as f:
            for index in range(len(self)):
                yield self.__item(f, index)


def save_as_edf(calib: Diffabs,
                basedir: Text) -> None:
    """Save the multi calib images into edf files in order to do the first
    calibration"""
    for idx, metadata in zip(calib.idxs, calib.frames()):
        base = os.path.splitext(os.path.basename(calib.filename))[0]
        output = os.path.join(basedir, base + '_%d.edf' % (idx,))
        edfimage(metadata.spectrum).write(output)


def get_wavelength(multicalib: Diffabs) -> float:
    """Return the wavelength"""
    return multicalib.wavelength


def get_calibrant(multicalib: Diffabs) -> pyFAI.calibrant.Calibrant:
    """Return the calibrant with the right wavelength"""
    calibrant = pyFAI.calibrant.get_calibrant(multicalib.calibrant)
    calibrant.wavelength = get_wavelength(multicalib)
    return calibrant


def get_detector(multicalib: Diffabs) -> pyFAI.Detector:
    return pyFAI.detector_factory(multicalib.detector)


def optimize_with_new_images(multicalib: Diffabs,
                             gonioref: GoniometerRefinement,
                             calibrant: pyFAI.calibrant.Calibrant,
                             pts_per_deg: float=1) -> None:
    """This function adds new images to the pool of data used for the
    refinement.  A set of new control points are extractred and a
    refinement step is performed at each iteration The last image of
    the serie is displayed

    """
    sg = None
    for idx, metadata in enumerate(multicalib.all_frames()):
        print()
        label = "CeO2_img%d_poni" % (idx,)
        if label in gonioref.single_geometries:
            continue
        print(label)

        sg = gonioref.new_geometry(label, image=metadata.image,
                                   metadata=metadata, calibrant=calibrant)
        print(sg.extract_cp(pts_per_deg=pts_per_deg))# , "watershed"))
    print("*"*50)
    gonioref.refine2()
    if sg:
        sg.geometry_refinement.set_param(gonioref.get_ai(sg.get_position()).param)  # noqa
        jupyter.display(sg=sg)


def calibration(json: str, multicalib:Diffabs) -> None:
    """Do a calibration with a bunch of images"""

    # Definition of the geometry refinement: the parameter order is
    # the same as the param_names
    calibrant = get_calibrant(multicalib)
    detector = get_detector(multicalib)

    distance = 0.39
    poni1 = -0.01
    poni2 = 0.033
    rot1 = 0
    rot2_scale = -1 * pi / 180.
    rot2_offset = radians(16)
    rot3 = radians(-90)

    parameters = [Parameter("dist", distance, (distance, distance)),
                  Parameter("poni1", poni1, (poni1-1, poni1+1)),
                  Parameter("poni2", poni2, (poni2-1, poni2+1)),
                  Parameter("rot1", rot1, (rot1, rot1)),
                  Parameter("rot2_scale", rot2_scale, (rot2_scale, rot2_scale)),
                  Parameter("rot2_offset", rot2_offset, (rot2_offset-0.1, rot2_offset+0.1)),
                  Parameter("rot3", rot3, (rot3, rot3))]

    params = {p.name: p.value for p in parameters}
    bounds = {p.name: p.bounds for p in parameters}
    param_names = [p.name for p in parameters]

    # Let's refine poni1 and poni2 also as function of the distance:

    trans_function = GeometryTransformation(param_names=param_names,
                                            pos_names=["delta"],
                                            dist_expr="dist",
                                            poni1_expr="poni1",
                                            poni2_expr="poni2",
                                            rot1_expr="rot1",
                                            rot2_expr="delta * rot2_scale + rot2_offset",
                                            rot3_expr="rot3")

    def pos_function(metadata: MetaData) -> Tuple[float]:
        """Definition of the function reading the detector position from the
        header of the image."""
        return metadata.delta,

    gonioref = GoniometerRefinement(params,  # initial guess
                                    bounds=bounds,
                                    pos_function=pos_function,
                                    trans_function=trans_function,
                                    detector=detector,
                                    wavelength=multicalib.wavelength)

    print("Empty refinement object:")
    print(gonioref)

    # Let's populate the goniometer refinement object with the known poni
    for idx, metadata in zip(multicalib.idxs, multicalib.frames()):
        label = "CeO2_img%d_poni" % (idx,)
        control_points = os.path.join(multicalib.basedir, "img%d_pts.npt" % (idx,))
        ai = pyFAI.load(os.path.join(multicalib.basedir, label))  # noqa
        print(ai)

        gonioref.new_geometry(label, metadata.image, metadata,
                              control_points, calibrant, ai)

    print("Filled refinement object:")
    print(gonioref)
    print(os.linesep + "\tlabel \t delta")
    for k, v in gonioref.single_geometries.items():
        print(k, v.get_position())

    for g in gonioref.single_geometries.values():
        ai = gonioref.get_ai(g.get_position())
        print(ai)

    for sg in gonioref.single_geometries.values():
        jupyter.display(sg=sg)

    gonioref.refine2()

    for multi in [multicalib]:
        optimize_with_new_images(multi, gonioref, calibrant)

    for idx, sg in enumerate(gonioref.single_geometries.values()):
        sg.geometry_refinement.set_param(gonioref.get_ai(sg.get_position()).param)
        jupyter.display(sg=sg)

    gonioref.save(json)
    pylab.show()


def integrate(json: Text) -> None:
    """Integrate a file with a json calibration file"""
    gonio = pyFAI.goniometer.Goniometer.sload(json)
    filename = os.path.join(ROOT, "tth2C-16_11_15-55-22_008.nxs")
    wavelength = 0.67186e-10
    multicalib = Diffabs(filename,
                        MetaDataSource(H5PathWithAttribute("interpretation", b"spectrum"),
                                       H5PathContains("scan_data/actuator_1_1")),
                        [], 0, "LaB6", "mythen", wavelength)

    # print(len(multicalib))
    # metadata = multicalib.__item__(21)

    # ai = gonio.get_ai(metadata.tth)
    # res = ai.integrate1d(metadata.spectrum, 1280, unit="2th_deg")
    # jupyter.plot1d(res)

    images = []
    positions = []
    for metadata in multicalib.all_frames():
        images.append(metadata.spectrum)
        positions.append((metadata.tth,))
    mai = gonio.get_mg(positions)
    res = mai.integrate1d(images, 30000)
    jupyter.plot1d(res)
    pylab.show()


def main() -> None:
    json = os.path.join(CALIB, "diffabs.json")
    filename = os.path.join(ROOT, "2017", "Run5", "2017-11-15", "scan_81.nxs")
    wavelength = 6.81198566418e-11
    calib = Diffabs(filename,
                    CALIB,
                    MetaDataSource(H5PathWithAttribute("interpretation", b"image"),
                                   H5PathContains("scan_data/actuator_1_1")),
                    [3, 4], "CeO2", "imxpad_s140", wavelength)

    calibration(json, calib)

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import math
import numpy

from gi.repository import GLib
from gi.repository import Hkl

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt

detector = Hkl.Detector.factory_new(Hkl.DetectorType(0))
detector.idx_set(1)

config = Hkl.geometry_factory_get_config_from_type(
    Hkl.GeometryType.KAPPA6C)
geometry = Hkl.Geometry.factory_newv(config, [math.radians(50.)])
delta = geometry.axes()[5]
# delta.parameter.range.min = 0
# values_w = [0., -60, 0., 90., 0., 60.]
values_w = [0., 120, 0., -90., 0., 60.]
geometry.set_axis_values_unit(values_w)
axis_names = [axis.name for axis in geometry.axes()]

sample = Hkl.Sample.new("toto")
lattice = sample.lattice_get()
lattice.set(1.54, 1.54, 1.54,
            math.radians(90.0),
            math.radians(90.0),
            math.radians(90.))
sample.lattice_set(lattice)

# get all engines for a given configuration
engines = Hkl.EngineList.factory(config)

# prepare the engines to work with the related geometry, detector and
# sample
engines.init(geometry, detector, sample)

#[0, 0, 1] -> [0, 1, 1]
n = 10
hkl0 = [0, 0, 1]
hkl1 = [0, 1, 1]
h = numpy.linspace(hkl0[0], hkl1[0], n + 1)
k = numpy.linspace(hkl0[1], hkl1[1], n + 1)
l = numpy.linspace(hkl0[2], hkl1[2], n + 1)

# get the hkl engine
hkl = engines.get_by_name("hkl")
eulerians = engines.get_by_name("eulerians")
pseudo_names = [
    pseudo_axis.parameter.name for pseudo_axis in eulerians.pseudo_axes()]

# set the hkl engine and get the results
trajectories = []

for idx, h, k, l in zip(range(n), h, k, l):
    try:
        hkl.set_values_unit([h, k, l])
        # print geometry.get_axis_values_unit()
        if engines.geometries.len != 4:
            print idx, h, k, l
        for i, item in enumerate(engines.geometries().items()):
            try:
                trajectories[i]
            except IndexError:
                trajectories.append([])
            values = item.geometry.get_axis_values_unit()
            print values, geometry.distance(item.geometry)
            trajectories[i].append(values)
        print
    except GLib.GError, err:
        pass

plt.plot(trajectories[0])
plt.ylim(-180, 180)


def full():
    plt.figure()
    maximum = min(4, len(trajectories))
    for i, trajectory in enumerate(trajectories):
        if i < maximum:
            plt.subplot(1, maximum, i + 1)
            plt.plot(trajectory)
            plt.ylim(-180, 180)



import bpy
import numpy
import pyFAI
import math
import functools
import random
import bmesh


SCALE = 100

# Definition du détecteur Cirpad

class Cirpad(pyFAI.detectors.Detector): # (Detector):
    MAX_SHAPE = (11200, 120)
    IS_FLAT = False
    IS_CONTIGUOUS = False
    force_pixel = True
    uniform_pixel = False
    aliases = ["XCirpad"]
    MEDIUM_MODULE_SIZE = (560, 120)
    MODULE_SIZE = (80, 120)  # number of pixels per module (y, x)
    PIXEL_SIZE = (130e-6, 130e-6)
    DIFFERENT_PIXEL_SIZE = 2.5
    rot = [0,0,-6.74]

    def _calc_pixels_size(self,length, module_size, pixel_size):
        size = numpy.ones(length)
        n = (length // module_size)
        for i in range(1, n):
            size[i * module_size - 1] = self.DIFFERENT_PIXEL_SIZE
            size[i * module_size] = self.DIFFERENT_PIXEL_SIZE
        return pixel_size * size

    def passage(self,corners,rot): #,u)  # A définir plus précisément comment on obtient rot et u
        shape = corners.shape
        origine = corners[0][0][0,:]
        nmd = rotation(corners,rot)
        u = corners[shape[0]-1][0][1,:] - corners[0][0][0,:]
        u = u / numpy.linalg.norm(u)
        s = rotation(u,rot)
        s = s / numpy.linalg.norm(s)
        v = numpy.array([-u[1],u[0],u[2]])
        r = origine - nmd[0][0][0,:]
        w = (0.1e-3+0.24e-3+75.14e-3)*u + (0.8e-3)*v + (0.55e-3)*s + r
        nmd = translation(nmd,w)
        return(nmd)

    def get_pixel_corners(self):
        pixel_size1 = self._calc_pixels_size(self.MEDIUM_MODULE_SIZE[0], self.MODULE_SIZE[0], self.PIXEL_SIZE[0])
        pixel_size2 = (numpy.ones(self.MEDIUM_MODULE_SIZE[1]) * self.PIXEL_SIZE[1]).astype(numpy.float32)
        # half pixel offset
        pixel_center1 = pixel_size1 / 2.0  # half pixel offset
        pixel_center2 = pixel_size2 / 2.0
        # size of all preceeding pixels
        pixel_center1[1:] += numpy.cumsum(pixel_size1[:-1])
        pixel_center2[1:] += numpy.cumsum(pixel_size2[:-1])

        pixel_center1.shape = -1, 1
        pixel_center1.strides = pixel_center1.strides[0], 0

        pixel_center2.shape = 1, -1
        pixel_center2.strides = 0, pixel_center2.strides[1]

        pixel_size1.shape = -1, 1
        pixel_size1.strides = pixel_size1.strides[0], 0

        pixel_size2.shape = 1, -1
        pixel_size2.strides = 0, pixel_size2.strides[1]

        # On calcule la position du premier module
        corners = numpy.zeros((self.MEDIUM_MODULE_SIZE[0], self.MEDIUM_MODULE_SIZE[1], 4, 3), dtype=numpy.float32)
        corners[:, :, 0, 1] = pixel_center1 - pixel_size1 / 2.0
        corners[:, :, 0, 2] = pixel_center2 - pixel_size2 / 2.0
        corners[:, :, 1, 1] = pixel_center1 + pixel_size1 / 2.0
        corners[:, :, 1, 2] = pixel_center2 - pixel_size2 / 2.0
        corners[:, :, 2, 1] = pixel_center1 + pixel_size1 / 2.0
        corners[:, :, 2, 2] = pixel_center2 + pixel_size2 / 2.0
        corners[:, :, 3, 1] = pixel_center1 - pixel_size1 / 2.0
        corners[:, :, 3, 2] = pixel_center2 + pixel_size2 / 2.0

        n_corners = corners
        # Puis on calcule les coins pour les 19 modules restant
        for i in range(1, 20):
            n_corners = self.passage(n_corners, self.rot)
            corners = numpy.concatenate((corners, n_corners), axis=0) # A voir la disposition finale souhaité
        return corners

    # Pas fait encore
    def calc_cartesian_positions(self,d1 = None, d2 = None, center = True, use_cython = True):
        if (d1 is None) or d2 is None:
            d1 = pyFAI.utils.expand2d(numpy.arange(self.MAX_SHAPE[0]).astype(numpy.float32), self.MAX_SHAPE[1], False)
            d2 = pyFAI.utils.expand2d(numpy.arange(self.MAX_SHAPE[1]).astype(numpy.float32), self.MAX_SHAPE[0], True)
        corners = self.get_pixel_corners()
        if center:
            # avoid += It modifies in place and segfaults
            d1 = d1 + 0.5
            d2 = d2 + 0.5
        if  False and use_cython:
            p1, p2, p3 = bilinear.calc_cartesian_positions(d1.ravel(), d2.ravel(), corners, is_flat=False)
            p1.shape = d1.shape
            p2.shape = d2.shape
            p3.shape = d2.shape
        else: # A verifier
            i1 = d1.astype(int).clip(0, corners.shape[0] - 1)
            i2 = d2.astype(int).clip(0, corners.shape[1] - 1)
            delta1 = d1 - i1
            delta2 = d2 - i2
            pixels = corners[i1, i2]
            if pixels.ndim == 3:
                A0 = pixels[:, 0, 0]
                A1 = pixels[:, 0, 1]
                A2 = pixels[:, 0, 2]
                B0 = pixels[:, 1, 0]
                B1 = pixels[:, 1, 1]
                B2 = pixels[:, 1, 2]
                C0 = pixels[:, 2, 0]
                C1 = pixels[:, 2, 1]
                C2 = pixels[:, 2, 2]
                D0 = pixels[:, 3, 0]
                D1 = pixels[:, 3, 1]
                D2 = pixels[:, 3, 2]
            else:
                A0 = pixels[:, :, 0, 0]
                A1 = pixels[:, :, 0, 1]
                A2 = pixels[:, :, 0, 2]
                B0 = pixels[:, :, 1, 0]
                B1 = pixels[:, :, 1, 1]
                B2 = pixels[:, :, 1, 2]
                C0 = pixels[:, :, 2, 0]
                C1 = pixels[:, :, 2, 1]
                C2 = pixels[:, :, 2, 2]
                D0 = pixels[:, :, 3, 0]
                D1 = pixels[:, :, 3, 1]
                D2 = pixels[:, :, 3, 2]

            # points A and D are on the same dim1 (Y), they differ in dim2 (X)
            # points B and C are on the same dim1 (Y), they differ in dim2 (X)
            # points A and B are on the same dim2 (X), they differ in dim1 (Y)
            # points C and D are on the same dim2 (X), they differ in dim1 (
            p1 = A1 * (1.0 - delta1) * (1.0 - delta2) \
                 + B1 * delta1 * (1.0 - delta2) \
                 + C1 * delta1 * delta2 \
                 + D1 * (1.0 - delta1) * delta2
            p2 = A2 * (1.0 - delta1) * (1.0 - delta2) \
                 + B2 * delta1 * (1.0 - delta2) \
                 + C2 * delta1 * delta2 \
                 + D2 * (1.0 - delta1) * delta2
            p3 = A0 * (1.0 - delta1) * (1.0 - delta2) \
                 + B0 * delta1 * (1.0 - delta2) \
                 + C0 * delta1 * delta2 \
                 + D0 * (1.0 - delta1) * delta2
            # To ensure numerical consitency with cython procedure.
            p1 = p1.astype(numpy.float32)
            p2 = p2.astype(numpy.float32)
            p3 = p3.astype(numpy.float32)
        return p1, p2, p3

# Fonction utilisée pour définir le Cirpad mais je ne sais pas si je les met dans la classe ou pas car elles sont banals

def M(theta, u):
    """
    :param theta: the axis value in radian
    :type theta: float
    :param u: the axis vector [x, y, z]
    :type u: [float, float, float]
    :return: the rotation matrix
    :rtype: numpy.ndarray (3, 3)
    """
    c = math.cos(theta)
    one_minus_c = 1 - c
    s = math.sin(theta)
    return [[c + u[0]**2 * one_minus_c,
             u[0] * u[1] * one_minus_c - u[2] * s,
             u[0] * u[2] * one_minus_c + u[1] * s],
            [u[0] * u[1] * one_minus_c + u[2] * s,
             c + u[1]**2 * one_minus_c,
             u[1] * u[2] * one_minus_c - u[0] * s],
            [u[0] * u[2] * one_minus_c - u[1] * s,
             u[1] * u[2] * one_minus_c + u[0] * s,
             c + u[2]**2 * one_minus_c]]


def rotation(md,rot):
    shape = md.shape
    axe = numpy.array([[1,0,0],[0,1,0],[0,0,1]])    # A voir si ce n'est pas une entrée
    P = functools.reduce(numpy.dot, [M(numpy.radians(rot[i]),axe[i]) for i in range(len(rot))])
    try :
        nmd = numpy.transpose(numpy.reshape(numpy.tensordot(P,numpy.reshape(numpy.transpose(md),(3,shape[0]*shape[1]*4)), axes=1),(3,4,shape[1],shape[0])))
    except IndexError :
        nmd =  numpy.transpose(numpy.tensordot(P,numpy.transpose(md), axes=1))
    return(nmd)


def translation(md, u):
    nmd = md + u
    return(nmd)

# Change coordonée pour passer de celle de pyFAI à celle de l'affichage.

def inter2(array):
    ar = numpy.empty(numpy.shape(array))
    ar[0] = array[2]
    ar[1] = array[1]
    ar[2] = array[0]
    return(ar)

# Affiche points  coins pixels module
def print_corners(corners, name):
    shape = corners.shape
    vertices = []
    vertices = numpy.reshape(corners, (shape[0] * shape [1] * shape[2], 3))
    contour_mesh = bpy.data.meshes.new("contour_mesh")
    contour_mesh .from_pydata(vertices, [], [])
    contour_mesh .update()
    contour_objet = bpy.data.objects.new(name, contour_mesh )
    scene = bpy.context.scene
    contour_objet.active_material = mat1
    scene.objects.link(contour_objet)

    bpy.context.scene.objects.active = bpy.context.scene.objects[name]
    bpy.ops.object.mode_set(mode = 'EDIT')
    bpy.ops.mesh.remove_doubles()
    bpy.ops.object.mode_set(mode = 'OBJECT')



# Materiau classique utilisé
mat1 = bpy.data.materials.new('Mat1')
mat1.diffuse_color = (1,1,1)

mat2 = bpy.data.materials.new('Mat2')
mat2.diffuse_color = (0,1,0)


# Affiche points centres pixels
def print_centers(centers, name):
    shape = centers[0].shape
    m = numpy.zeros((3, shape[0], shape[1]))
    for i in range(3):
        if centers[i] is not None:
            m[i] = centers[i] * SCALE

    m = inter2(m)

    # Creation des vertices
    vertices = numpy.reshape(numpy.transpose(m), (shape [0] * shape[1], 3))

    # Creation du mesh à partir des vertices
    contour_mesh = bpy.data.meshes.new("contour_mesh")
    contour_mesh.from_pydata(vertices, [], [])
    contour_mesh.update()

    # Creation d'un object avec ce mesh
    contour_objet = bpy.data.objects.new(name, contour_mesh)

    # si pas de materiaux, le menu n'apparait pas ???
    contour_objet.active_material = mat2

    # ajout de l'objet à la scène pour qu'il devienne visible
    scene = bpy.context.scene
    scene.objects.link(contour_objet)


# Affiche le detecteur avec ses couleurs, pas tourner dans le bon sens pour l'instant
def print_pixel(corners, image):

    shape = corners.shape

    # Creer des vertices du detecteur
    vertices = numpy.reshape(corners * SCALE, (shape[0] * shape [1] * shape[2], 3))

    # creation de la list contenant les indices des vertices formant une face (un pixel)
    faces = [[4*i,4*i+1,4*i+2,4*i+3] for i in range(shape[0]*shape[1])]

    # creation du mesh avec ces vertices et ces faces.
    contour_mesh = bpy.data.meshes.new("contour_mesh")
    contour_mesh.from_pydata(vertices, [], faces)
    contour_mesh.update()

    # creation de l'object a partir du mesh
    contour_objet = bpy.data.objects.new('detector', contour_mesh )

    contour_objet.active_material = mat1

    # ajout de l'object dans la scene.
    scene = bpy.context.scene
    scene.objects.link(contour_objet)

    # Ajoute les couleurs
    # on cree un groupe de vertex par couleurs de la LUT
    # ici la lut est simple 0-255 (sur le rouge)
    max = int(image.max()) + 1
    MyVertices = [[] for i in range(max)]

    # On partour toute l'image et pour chaque intensité, on rajoute les index des
    # vertex de la list vertices formant une face dans le bon groupe de vertex.
    # TODO improve speed :)
    width = shape[1]

    # TODO voir comment utiliser plutot les faces pour ces couleurs.
    for i in range(shape[0]):
        i_width = i * width
        for j in range(shape[1]):
            i_width_j = 4 * (i_width + j)
            l = range(i_width_j, i_width_j + 4)
            MyVertices[int(image[i,j])].extend(l)
    bpy.context.scene.objects.active = bpy.context.scene.objects['detector']

    # Met les couleurs par groupe
    for k in range(max):

        name = "MyVertexGroup"+str(k)

         #Create a Vertex Group
        Group1 = bpy.context.object.vertex_groups.new(name)

        #Add the vertices to the vertex group
        #with Weight = 1.0 The Weight isn't
        #relevant in this case
        Group1.add(MyVertices[k], 1.0, 'ADD')

        #Select the vertex group
        bpy.ops.object.vertex_group_set_active(group= name)

        bpy.ops.object.material_slot_add()
        bpy.context.object.material_slots[bpy.context.object.material_slots.__len__() - 1].material = MakeMaterial_1(k,max)
        bpy.ops.object.mode_set(mode = 'EDIT')
        #Deselect all the vertices
        bpy.ops.mesh.select_all(action='DESELECT')
        #Select the vertices of the vertex group
        bpy.ops.object.vertex_group_select()
        #Assign the material on the selected vertices
        bpy.ops.object.material_slot_assign()
        bpy.ops.object.mode_set(mode = 'OBJECT')

    bpy.ops.object.mode_set(mode = 'EDIT')
    bpy.ops.mesh.remove_doubles()
    bpy.ops.object.mode_set(mode = 'OBJECT')


# Fonction servant a créé un materiau par couleur, dépend de l'intensité
def MakeMaterial_1(k, max):
    mat = bpy.data.materials.new("Mat2")
    mat.diffuse_shader = 'MINNAERT'
    mat.diffuse_color = (1, round(1-k/max,4), round(1-k/max,4))
    mat.darkness = 0.8
    return mat

detector = pyFAI.detectors.ImXPadS70()
corners = detector.get_pixel_corners()
centers = detector.calc_cartesian_positions()

shape = corners.shape[0:2]

image = numpy.random.randint(256, size=shape)

print_pixel(corners,image)



# -*- encoding: utf-8 -*-
"""
This file is part of the hkl library.

The hkl library is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The hkl library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the hkl library.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2011-2013 Synchrotron SOLEIL
                        L'Orme des Merisiers Saint-Aubin
                        BP 48 91192 GIF-sur-YVETTE CEDEX
Authors: Picca Frédéric-Emmanuel <picca@synchrotron-soleil.fr>
"""
import math
from gi.repository import Hkl

"""
static int test_engine(struct hkl_test *test,
		       HklEngine *engine, HklGeometry *geometry,
		       HklDetector *detector, HklSample *sample)
{
	size_t i, j, k, f_idx;
	double *values = alloca(HKL_LIST_LEN(engine->pseudoAxes) * sizeof(*values));
	int miss = 0;

	/* randomize the geometry */
	hkl_geometry_randomize(geometry);
	
	for(f_idx=0; f_idx<HKL_LIST_LEN(engine->modes); ++f_idx) {
		hkl_engine_current_mode_set(engine, f_idx);
		/* for now unactive the eulerians check */
		if(!strcmp(engine->mode->name, "eulerians"))
			continue;
		miss = 0;
		for(i=0;i<N;++i) {
			int res;
			size_t len = HKL_LIST_LEN(engine->pseudoAxes);

			/* randomize the pseudoAxes values */
			for(j=0; j<len; ++j) {
				HklParameter *parameter = (HklParameter *)(engine->pseudoAxes[j]);
				hkl_parameter_randomize(parameter);

				values[j] = parameter->value;
			}

			/* randomize the parameters */
			for(j=0; j<HKL_LIST_LEN(engine->mode->parameters); ++j)
				hkl_parameter_randomize(&engine->mode->parameters[j]);

			/* pseudo -> geometry */
			hkl_engine_initialize(engine, NULL);
			/* hkl_engine_fprintf(stderr, engine); */
			res = hkl_engine_set(engine, NULL);

			/* geometry -> pseudo */
			if (res == HKL_SUCCESS) {
				size_t g_len = hkl_geometry_list_len(engine->engines->geometries);
				/* check all finded geometries */
				/* hkl_engine_fprintf(stderr, engine); */

				for(j=0; j<g_len; ++j) {
					/* first modify the pseudoAxes values */
					/* to be sure that the result is the */
					/* computed result. */
					for(k=0; k<len; ++k)
						((HklParameter *)engine->pseudoAxes[k])->value = 0.;

					hkl_geometry_init_geometry(geometry,
								   engine->engines->geometries->items[j]->geometry);
					hkl_engine_get(engine, NULL);

					for(k=0; k<len; ++k) {
						HKL_ASSERT_DOUBLES_EQUAL(values[k],
									 ((HklParameter *)engine->pseudoAxes[k])->value,
									 HKL_EPSILON);
					}
				}
			} else
				miss++;
		}

#if with_log
		fprintf(stderr, "\n\"%s\" \"%s\" missed : %d",
			engine->geometry->name,
			engine->mode->name, miss);
#endif

	}
	
#if with_log
	fprintf(stderr, "\n");
#endif

	return HKL_TEST_PASS;
}

#define test_engines(test, engines) do{					\
		size_t i;						\
		for(i=0; i<HKL_LIST_LEN(engines->engines); ++i){	\
			if (!test_engine(test, engines->engines[i],	\
					 engines->geometry,		\
					 engines->detector,		\
					 engines->sample))		\
				return HKL_TEST_FAIL;			\
		}							\
	}while(0)
"""


def test_all():
    detector = Hkl.Detector().factory_new(getattr(Hkl.DetectorType, '0D'))

    # attache to the second holder
    detector.idx = 1

    # create the right diffractometer geometry
    config = Hkl.Geometry.factory_get_config_from_type(Hkl.GeometryType.KAPPA6C)
    geometry = Hkl.Geometry.factory_newv(config, [50. * math.pi / 180.])
    geometry.source.wave_length = 1.54

    # configure the sample
    sample = Hkl.Sample.new("toto")
    sample.set_lattice(1.54, 1.54, 1.54, 90., 90., 90.)

    # create the Engines
    engines = Hkl.engine_list_factory(config)
    for engine in engines.engines:
        print engine

test_all()
"""
	const HklGeometryConfig *config;
	HklGeometry *geometry = NULL;
	HklDetector *detector = hkl_detector_factory_new(HKL_DETECTOR_TYPE_0D);
	HklSample *sample = hkl_sample_new("test", HKL_SAMPLE_TYPE_MONOCRYSTAL);
	HklEngineList *engines;

	/* attach to the second holder */
	detector->idx = 1;

	/* test all E4CV engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_EULERIAN4C_VERTICAL);
	geometry = hkl_geometry_factory_new(config);
	engines = hkl_engine_list_factory(config);
	hkl_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_engine_list_free(engines);

	/* test all E6C HKL engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_EULERIAN6C);
	geometry = hkl_geometry_factory_new(config);
	engines = hkl_engine_list_factory(config);
	hkl_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_engine_list_free(engines);

	/* test all K4CV HKL engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_KAPPA4C_VERTICAL);
	geometry = hkl_geometry_factory_new(config, 50 * HKL_DEGTORAD);
	engines = hkl_engine_list_factory(config);
	hkl_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_engine_list_free(engines);

	/* test all K6C engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_KAPPA6C);
	geometry = hkl_geometry_factory_new(config, 50 * HKL_DEGTORAD);
	engines = hkl_engine_list_factory(config);
	hkl_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_engine_list_free(engines);

	/* test all ZAXIS engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_ZAXIS);
	geometry = hkl_geometry_factory_new(config);
	engines = hkl_engine_list_factory(config);
	hkl_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_engine_list_free(engines);

	/* test all SOLEIL SIXS MED 2+2 engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_SOLEIL_SIXS_MED_2_2);
	geometry = hkl_geometry_factory_new(config);
	engines = hkl_engine_list_factory(config);
	hkl_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_engine_list_free(engines);

	/* test all SOLEIL MARS engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_SOLEIL_MARS);
	geometry = hkl_geometry_factory_new(config);
	engines = hkl_engine_list_factory(config);
	hkl_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_engine_list_free(engines);

	/* test all SOLEIL SIXS MED 1+2 engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_SOLEIL_SIXS_MED_1_2);
	geometry = hkl_geometry_factory_new(config);
	engines = hkl_engine_list_factory(config);
	hkl_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_engine_list_free(engines);

	/* test all PETRA3 P09 EH2 engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_PETRA3_P09_EH2);
	geometry = hkl_geometry_factory_new(config);
	engines = hkl_engine_list_factory(config);
	hkl_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_engine_list_free(engines);

	/* test all SOLEIL SIXS MED 2+3 engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_SOLEIL_SIXS_MED_2_3);
	geometry = hkl_geometry_factory_new(config);
	engines = hkl_pseudo_axis_engine_list_factory(config);
	hkl_pseudo_axis_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_pseudo_axis_engine_list_free(engines);

	/* test all SOLEIL SIRIUS TURRET engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_SOLEIL_SIRIUS_TURRET);
	geometry = hkl_geometry_factory_new(config);
	engines = hkl_pseudo_axis_engine_list_factory(config);
	hkl_pseudo_axis_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_pseudo_axis_engine_list_free(engines);

	/* test all SOLEIL SIRIUS KAPPA engines */
	config = hkl_geometry_factory_get_config_from_type(HKL_GEOMETRY_TYPE_SOLEIL_SIRIUS_KAPPA);
	geometry = hkl_geometry_factory_new(config, 50 * HKL_DEGTORAD);
	engines = hkl_pseudo_axis_engine_list_factory(config);
	hkl_pseudo_axis_engine_list_init(engines, geometry, detector, sample);
	test_engines(test, engines);
	hkl_geometry_free(geometry);
	hkl_pseudo_axis_engine_list_free(engines);

	hkl_detector_free(detector);
	hkl_sample_free(sample);

	return HKL_TEST_PASS;
}
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
This file is part of the hkl library.

The hkl library is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

The hkl library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with the hkl library.  If not, see <http://www.gnu.org/licenses/>.

Copyright (C) 2003-2012, 2019 Synchrotron SOLEIL
                        L'Orme des Merisiers Saint-Aubin
                        BP 48 91192 GIF-sur-YVETTE CEDEX
Authors: Picca Frédéric-Emmanuel <picca@synchrotron-soleil.fr>
"""

import math
from gi.repository import GLib
from gi.repository import Hkl

detector = Hkl.Detector.factory_new(Hkl.DetectorType(0))

factory = Hkl.factories()['K6C']
geometry = factory.create_new_geometry()
values_w = [0., 30., 0., 0., 0., 60.]
geometry.axis_values_set(values_w, Hkl.UnitEnum.USER)
axis_names = geometry.axis_names_get()
print(geometry.name_get(), "diffractometer has", len(axis_names),\
      "axes : ", axis_names)
print(values_w)

sample = Hkl.Sample.new("toto")
lattice = Hkl.Lattice.new(1.54, 1.54, 1.54,
                          math.radians(90.0),
                          math.radians(90.0),
                          math.radians(90.))
sample.lattice_set(lattice)

# compute all the pseudo axes managed by all engines
engines = factory.create_new_engine_list()
engines.init(geometry, detector, sample)
engines.get()

# get the hkl engine and do a computation
hkl = engines.engine_get_by_name("hkl")
values = hkl.pseudo_axis_values_get(Hkl.UnitEnum.USER)
print("read : ", values)

# set the hkl engine and get the results
for _ in range(100):
    try:
        print()
        solutions = hkl.pseudo_axis_values_set(values,
                                               Hkl.UnitEnum.USER)
        print(hkl.pseudo_axis_values_get(Hkl.UnitEnum.USER))

        print("idx".center(15)),
        for name in axis_names:
            print("{}".format(name.center(15))),
        print()

        for i, item in enumerate(solutions.items()):
            read = item.geometry_get().axis_values_get(Hkl.UnitEnum.USER)
            print("{}".format(repr(i).center(15))),
            for value in read:
                print("{}".format(repr(value)[:15].center(15))),
            print()
    except GLib.GError as err:
        print(values, err)
    values[1] += .01



#!/usr/bin/env python3

import os
import math
import unittest

from collections import namedtuple
from gi.repository import GLib
import gi
gi.require_version("Hkl", "5.0")
from gi.repository import Hkl
from numpy import (array, cross, dot, empty, hstack, reshape, vstack)
from numpy.linalg import inv, norm

#########
# Types #
#########

Detector = namedtuple('Detector', ['type'])
Diffractometer = namedtuple('Diffractometer', ['dtype', 'sample',
                                               'detector', 'source', 'engine'])
Engine = namedtuple('Engine', ['name', 'mode'])
Lattice = namedtuple('Lattice', ['a', 'b', 'c', 'alpha', 'beta', 'gamma'])
Reflection = namedtuple('Reflection', ['hkl', 'values'])
Sample = namedtuple('Sample', ['lattice', 'or0', 'or1', 'ux', 'uy', 'uz'])
Source = namedtuple('Source', ['wavelength', 'energy'])

HklDiffractometer = namedtuple('HklDiffractometer', ['sample', 'detector',
                                                     'geometry', 'engines'])


##################
# Helper methods #
##################

def hkl_matrix_to_numpy(m):
    M = empty((3, 3))
    for i in range(3):
        for j in range(3):
            M[i, j] = m.get(i, j)
    return M


def from_numpy_to_hkl_vector(v):
    V = Hkl.Vector()
    V.init(v[0], v[1], v[2])
    return V


#########################
# Hkl Type constructors #
#########################

def new_hkl_sample(config):
    # sample
    sample = Hkl.Sample.new("test")
    lattice = Hkl.Lattice.new(config.lattice.a,
                              config.lattice.b,
                              config.lattice.c,
                              math.radians(config.lattice.alpha),
                              math.radians(config.lattice.beta),
                              math.radians(config.lattice.gamma))
    sample.lattice_set(lattice)

    parameter = sample.ux_get()
    parameter.value_set(config.ux, Hkl.UnitEnum.USER)
    sample.ux_set(parameter)

    parameter = sample.uy_get()
    parameter.value_set(config.uy, Hkl.UnitEnum.USER)
    sample.uy_set(parameter)

    parameter = sample.uz_get()
    parameter.value_set(config.uz, Hkl.UnitEnum.USER)
    sample.uz_set(parameter)

    return sample


def new_hkl_detector(config):
    return Hkl.Detector.factory_new(config.type)


def new_hkl_geometry(dtype, wavelength, init_values):
    factory = Hkl.factories()[dtype]
    geometry = factory.create_new_geometry()
    geometry.axis_values_set(init_values, Hkl.UnitEnum.USER)
    geometry.wavelength_set(wavelength, Hkl.UnitEnum.USER)
    return geometry


def new_hkl_engines(dtype):
    factory = Hkl.factories()[dtype]
    engines = factory.create_new_engine_list()
    return engines


def new_hkl_diffractometer(config):
    sample = new_hkl_sample(config.sample)

    detector = new_hkl_detector(config.detector)

    # add reflection or0
    geometry = new_hkl_geometry(config.dtype,
                                config.source.wavelength,
                                config.sample.or0.values)
    or0 = sample.add_reflection(geometry, detector,
                                config.sample.or0.hkl[0],
                                config.sample.or0.hkl[1],
                                config.sample.or0.hkl[2])

    # add reflection or1
    geometry.axis_values_set(config.sample.or1.values,
                             Hkl.UnitEnum.USER)
    or1 = sample.add_reflection(geometry, detector,
                                config.sample.or1.hkl[0],
                                config.sample.or1.hkl[1],
                                config.sample.or1.hkl[2])

    # compute UB with or0 and or1
    # sample.compute_UB_busing_levy(or0, or1)
    # UB = hkl_matrix_to_numpy(sample.UB_get())

    # compute angles for reciprocal lattice vector h, k, l
    engines = new_hkl_engines(config.dtype)

    # set the engine mode
    engine = engines.engine_get_by_name(config.engine.name)
    engine.current_mode_set(config.engine.mode)

    return HklDiffractometer(sample, geometry, detector, engines)

#################
# Config parser #
#################

def find(filename):
    if os.path.exists(filename):
        return filename
    else:
        datadir = os.getenv('DATADIR')
        if datadir:
            filename = os.path.join(datadir, filename)
            if os.path.exists(filename):
                return filename
            else:
                raise Exception("Can not find: " + filename)
        else:
            raise Exception("Cannot find: " + filename)


def parse_reflection(line):
    ref = line.split()
    hkl = [float(x) for x in ref[2:5]]
    angles = [float(x) for x in ref[7:13]]
    return Reflection(hkl, angles)


def parse(filename, dtype):
    with open(filename, 'r') as f:
        for line in f:
            if str(line).find('Wavelength') != -1:
                wavelength = float(line.split()[1])
                energy = 12.39842 / wavelength
            if str(line).find('A') != -1 and str(line).find('B') != -1 and str(line).find('C') != -1:
                abc = line.split()
                a = float(abc[1])
                b = float(abc[3])
                c = float(abc[5])
            if str(line).find('Alpha') != -1 and str(line).find('Beta') != -1 and str(line).find('Gamma') != -1:
                abg = line.split()
                alpha = float(abg[1])
                beta = float(abg[3])
                gamma = float(abg[5])
            if str(line).find('R0') != -1:
                or0 = parse_reflection(line)
            if str(line).find('R1') != -1:
                or1 = parse_reflection(line)
            if str(line).find('Ux') != -1 and str(line).find('Uy') != -1 and str(line).find('Uz') != -1:
                uxuyuz = line.split()
                ux = float(uxuyuz[1])
                uy = float(uxuyuz[3])
                uz = float(uxuyuz[5])
            if str(line).find('Engine') != -1:
                engine_name = line.split()[1]
            if str(line).find('Mode') != -1:
                mode_name = line.split()[1]

        lattice = Lattice(a, b, c, alpha, beta, gamma)
        sample = Sample(lattice, or0, or1, ux, uy, uz)
        detector = Detector(0)
        source = Source(wavelength, energy)
        engine = Engine(engine_name, mode_name)
        return Diffractometer(dtype, sample, detector, source, engine)


######################
# hight level method #
######################

def ca(config, hkl, engine_name='hkl'):
    sample, geometry, detector, engines = new_hkl_diffractometer(config)
    engines.init(geometry, detector, sample)
    engine = engines.engine_get_by_name(engine_name)
    solutions = engine.pseudo_axis_values_set(hkl,
                                              Hkl.UnitEnum.USER)
    first_solution = solutions.items()[0]
    values = first_solution.geometry_get().axis_values_get(Hkl.UnitEnum.USER)
    return Reflection(hkl, values)


def get_UB(config):
    sample, geometry, detector, engines = new_hkl_diffractometer(config)
    return hkl_matrix_to_numpy(sample.UB_get())


def get_R_and_P(config, values):
    sample, geometry, detector, engines = new_hkl_diffractometer(config)
    geometry.axis_values_set(values, Hkl.UnitEnum.USER)
    R = hkl_matrix_to_numpy(geometry.sample_rotation_get(sample).to_matrix())
    P = hkl_matrix_to_numpy(geometry.detector_rotation_get(detector).to_matrix())
    return R, P

##############
# Unit tests #
##############

class Polarisation(unittest.TestCase):
    def test_petraIII(self):
        dtype = "E6C"
        # RUBh = kf - ki = (P ki - ki) = (P - I) ki

        config = parse(find('crystal.ini'), dtype)
        print(config)

        gaga = ca(config, [0.5, 14.5, 0.43])
        print(gaga)

        UB = get_UB(config)
        print("UB: ")
        print(UB)

        # the hkl vectors a*, b*, c* expressed in the laboratory basis for the current Q
        # transformation matrix T reciprocal space --> laboratory
        # values_w = [0, 34.16414, 79.52420, 0, 0, 38.29633]
        # values_w = [0, 35.06068, 80.78517, 0, 0, 43.91934]
        # values_w = [0, 36.45961, 81.75533, 0, 0, 49.7139]
        # values_w = [0, 38.24551, 82.52447, 0, 0, 55.68957]
        # values_w = [0, 40.34955, 83.14900, 0, 0, 61.86603]
        # values_w = [0, 42.73321, 83.66607, 0, 0, 68.27333]
        # values_w = [0, 45.37981, 84.10117, 0, 0, 74.95312]
        # values_w = [0, 48.29079, 84.47230, 0, 0, 81.96226]
        values_w = [0, 51.48568, 84.79259, 0, 0, 89.37964]  # mu, omega, chi, phi, gamma, delta

        R, P = get_R_and_P(config, values_w)
        print("R: ")
        print(R)
        print("P: ")
        print(P)
        RUB = dot(R, UB)
        print("RUB: ")
        print(RUB)

        astar = dot(RUB, [1, 0, 0])
        bstar = dot(RUB, [0, 1, 0])
        cstar = dot(RUB, [0, 0, 1])

        # transformation matrix: reciprocal space --> laboratory
        T = hstack((reshape(astar, (3, 1)),
                    reshape(bstar, (3, 1)),
                    reshape(cstar, (3, 1))))
        Tbis = vstack((reshape(astar, (1, 3)),
                       reshape(bstar, (1, 3)),
                       reshape(cstar, (1, 3))))
        # transformation matrix: laboratory --> reciprocal space
        Tinv = inv(T)

        print('')
        # print 'cstar in laboratory frame       :',cstar
        # print 'cstar in laboratory frame from T:',dot(T, hkl)
        # print 'cstar in rec. space from Tinv   :',dot(Tinv, dot(T, hkl))

        # compute kf
        ki = array([1, 0, 0]) * math.pi * 2 / config.source.wavelength
        kf = dot(P, ki)

        # compute Q
        Q = kf - ki

        print('')
        print('Energy (keV):', config.source)
        print('Lattice parameters:', config.sample.lattice)
        print('1st orienting reflection:', config.sample.or0.hkl, 'with angles: ', config.sample.or0.values)
        print('2nd orienting reflection:', config.sample.or1.hkl, 'with angles: ', config.sample.or1.values)
        print('')

        print('UB matrix:')
        print(UB)
        print('')
        print('Transformation matrix T(reciprocal space)--> laboratory frame:')
        print(T,)
        print('')
        print('')
        print('Transformation matrix T(laboratory frame)--> reciprocal space :')
        print(Tinv)

        # compute Q
        # hkl = [0.5, 6.5, 0.43]
        # hkl = [0.5, 7.5, 0.43]
        # hkl = [0.5, 8.5, 0.43]
        # hkl = [0.5, 9.5, 0.43]
        # hkl = [0.5, 10.5, 0.43]
        # hkl = [0.5, 11.5, 0.43]
        # hkl = [0.5, 12.5, 0.43]
        # hkl = [0.5, 13.5, 0.43]
        hkl = [0.5, 14.5, 0.43]

        '''
        print ''
        print 'Q in lab. frame from code   :', dot(dot(R, UB), hkl), ', normalized:',(dot(dot(R, UB), hkl))/norm(dot(dot(R, UB), hkl)), ', norm:', norm(dot(dot(R, UB), hkl))
        print 'Q in lab. frame from T      :', dot(T, hkl), ', normalized:', (dot(T, hkl))/norm(dot(T, hkl)), ', norm:', norm(dot(T, hkl))
        print 'Q in lab. frame from ki, kf :', Q, ', normalized:', Q/norm(Q),', norm:', norm(Q)
        print ''
        print 'Q in rec. space from Tinv of T   :', dot(Tinv, dot(T, hkl))
        print ''
        print 'difference factor:',(norm(dot(T, hkl)))/(norm(Q))
        print ''
        print 'kf',kf,', norm:',norm(kf)
        print 'ki',ki,', norm:',norm(ki)
        '''
        print('')
        print('Q in rec. space from Tinv of T   :', dot(Tinv, dot(T, hkl)))
        print('')
        #
        # compute u1, u2, u3 in reciprocal space coordinates
        # u1,u2,u3 in laboratory frame
        u1xyz = ki+kf
        u2xyz = cross(ki, kf)
        u3xyz = ki-kf
        # print '(u1,u2,u3) in laboratory frame:',u1xyz,u2xyz,u3xyz
        # u1,u2,u3 in reciprocal space
        u1 = dot(Tinv, u1xyz) / norm(dot(Tinv, u1xyz))
        u2 = dot(Tinv, u2xyz) / norm(dot(Tinv, u2xyz))
        u3 = dot(Tinv, u3xyz) / norm(dot(Tinv, u3xyz))
        u1 = dot(Tinv, u1xyz)
        u2 = dot(Tinv, u2xyz)
        u3 = dot(Tinv, u3xyz)
        print('(u1,u2,u3) in reciprocal space from Tinv, unnormalized:',\
            u1, u2, u3)
        print('(u1,u2,u3) in reciprocal space from Tinv, normalized to 1:',\
            dot(Tinv, u1xyz) / norm(dot(Tinv, u1xyz)), \
            dot(Tinv, u2xyz) / norm(dot(Tinv, u2xyz)), \
            dot(Tinv, u3xyz) / norm(dot(Tinv, u3xyz)))
        # print '(u1,u2,u3) in reciprocal space from Tinv:', \
        #    dot(Tinv, u1xyz), dot(Tinv, u2xyz), dot(Tinv, u3xyz)
        print('')
        # TRANSFORMATION MATRIX reciprocal lattice basis to u1 u2 u3 basis
        ABC = hstack((reshape([1, 0, 0], (3, 1)),
                      reshape([0, 1, 0], (3, 1)),
                      reshape([0, 0, 1], (3, 1))))
        U = hstack((reshape(u1, (3, 1)),
                    reshape(u2, (3, 1)),
                    reshape(u3, (3, 1))))
        M = dot(ABC, inv(U))
        print('Transformation matrix reciprocal lattice basis to u1 u2 u3 basis:')
        print(M)

        # keep in order to pass the test
        self.assertTrue(True)



if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
""" coding: utf-8 """
"""
ceo2 :: XRDSample
ceo2 = XRDSample "CeO2"
       (published </> "xrd" </> "CeO2")
       [ XrdNxs bins multibins threshold skipedFrames (XrdSourceNxs n) | n <-
         [ mkNxs (project </> "2016" </> "Run4" </> "2016-09-07" </> "IHR_29.nxs") "scan_29" h5path'
         , mkNxs (project </> "2016" </> "Run4" </> "2016-09-07" </> "IHR_30.nxs") "scan_30" h5path'
         , mkNxs (project </> "2016" </> "Run4" </> "2016-09-07" </> "IHR_56.nxs") "scan_56" h5path'
         , mkNxs (project </> "2016" </> "Run4" </> "2016-09-07" </> "IHR_58.nxs") "scan_58" h5path'
         ]
       ]
"""
from typing import Iterator, List, Tuple

import os
import functools

from collections import namedtuple

import h5py
import numpy
import pylab
import pyFAI

from fabio.edfimage import edfimage

from pyFAI.goniometer import GeometryTransformation, GoniometerRefinement
from pyFAI.gui import jupyter


ROOT = "/nfs/ruche-diffabs/diffabs-soleil/com-diffabs/"
PUBLISHED = os.path.join(ROOT, "2016", "Run4B", "OutilsMetallo_CarolineHamon")

CALIB = os.path.join(ROOT, "2016", "Run4", "2016-09-07", "IHR_30.nxs")

# H5Path data
H5PathContains = namedtuple("H5PathContains", "path")
H5PathOptionalItemValue = namedtuple('H5OptionalItemValue', ['path', 'default'])

MetaDataSource = namedtuple("MetaDataSource", ["img", "tx", "tz"])

MetaData = namedtuple("MetaData", ["img", "tx", "tz"])

MultiCalib = namedtuple("MultiCalib", ["filename", "metasources", "idxs",
                                       "calibrant", "detector", "wavelength"])

Parameter = namedtuple("Parameter", ["name", "value", "bounds"])


def visit_attrs(acc: List[h5py.Dataset], _name: str, obj) -> None:
    """extract all the images and accumulate them in the acc variable"""
    if isinstance(obj, h5py.Dataset):
        if 'interpretation' in obj.attrs and obj.attrs['interpretation'] == b'image':
            acc.append(obj)
    return None


def get_images(h5file: h5py.File) -> List[h5py.Dataset]:
    acc = []  # type: List[h5py.Dataset]
    h5file.visititems(functools.partial(visit_attrs, acc))
    return acc


def visit_item(key: str, name: str, obj: h5py.Dataset) -> h5py.Dataset:
    if key in name:
        return obj


def get_item(h5file: h5py.File, item):
    res = None
    if isinstance(item, H5PathContains):
        res = h5file.visititems(functools.partial(visit_item, item.path))
    elif isinstance(item, H5PathOptionalItemValue):
        _item = h5file.visititems(functools.partial(visit_item, item.path))
        res = _item.value if _item else item.default
    return res


def get_metadata(h5file: h5py.File,
                 multicalib: MultiCalib,
                 index: int) -> MetaData:
    return MetaData(get_images(h5file)[0][index],
                    get_item(h5file, multicalib.metasources.tx)[index],
                    get_item(h5file, multicalib.metasources.tz))

def gen_metadata(h5file: h5py.File, multicalib: MultiCalib) -> Iterator[MetaData]:
    imgs = get_images(h5file)[0]
    for idx in range(imgs.shape[0]):
        yield get_metadata(h5file, multicalib, idx)

def save_as_edf(calib: MultiCalib, basedir: str) -> None:
    """Save the multi calib images into edf files in order to do the first
    calibration

    """
    with h5py.File(calib.filename, mode='r') as h5file:
        imgs = get_images(h5file)[0]
        for idx in calib.idxs:
            img = imgs[idx]
            base = os.path.splitext(os.path.basename(calib.filename))[0]
            output = os.path.join(basedir, base + '_%d.edf' % (idx,))
            edfimage(img).write(output)


def get_total_length(calib: MultiCalib) -> int:
    """Return the total number of frame of the calib file"""
    with h5py.File(calib.filename, mode='r') as f:
        return get_images(f)[0].shape[0]


def optimize_with_new_images(h5file: h5py.File,
                             multicalib: MultiCalib,
                             gonioref,
                             calibrant: pyFAI.calibrant.Calibrant,
                             pts_per_deg: float=1) -> None:
    """This function adds new images to the pool of data used for the
    refinement.  A set of new control points are extractred and a
    refinement step is performed at each iteration The last image of
    the serie is displayed

    """
    sg = None
    for idx in range(get_total_length(multicalib)-1):
        print()
        base = os.path.splitext(os.path.basename(multicalib.filename))[0]

        label = base + "_%d" % (idx,)
        if label in gonioref.single_geometries:
            continue
        print(label)
        metadata = get_metadata(h5file, multicalib, idx)
        sg = gonioref.new_geometry(label, image=metadata.img, metadata=metadata, calibrant=calibrant)
        print(sg.extract_cp(pts_per_deg=pts_per_deg))
    print("*"*50)
    gonioref.refine2()
    if sg:
        sg.geometry_refinement.set_param(gonioref.get_ai(sg.get_position()).param)
        jupyter.display(sg=sg)

# Extraction de l'image n°5 afin de faire la calibration avec pyFAI-calib2.

# In[5]:

# save this image as edf in order to generate the poni with pyFAI-calib2

def get_wavelength(multicalib: MultiCalib) -> float:
    """Return the wavelength"""
    return multicalib.wavelength


def get_calibrant(multicalib: MultiCalib) -> pyFAI.calibrant.Calibrant:
    """Return the calibrant with the right wavelength"""
    calibrant = pyFAI.calibrant.get_calibrant(multicalib.calibrant)
    calibrant.wavelength = get_wavelength(multicalib)
    return calibrant


def get_detector(multicalib: MultiCalib) -> pyFAI.Detector:
    return pyFAI.detector_factory(multicalib.detector)


def calibration(json: str) -> None:
    """Do a calibration with a bunch of images"""

    wavelength = 4.85945727522e-11

    multicalib = MultiCalib(os.path.join(ROOT, "scan_3_01.nxs"),
                            MetaDataSource("",
                                           H5PathContains("scan_data/actuator_1_1"),
                                           H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", 0.0)),
                            [2, 5, 8], "LaB6", "xpad_flat", wavelength)

    multicalib2 = MultiCalib(os.path.join(ROOT, "scan_4_01.nxs"),
                             MetaDataSource("",
                                            H5PathContains("scan_data/actuator_1_1"),
                                            H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", -1.0)),
                             [], "LaB6", "xpad_flat", wavelength)

    # save all the ref as images in order to do the calibration with
    # pyFAI-calib[2].
    save_as_edf(multicalib, PUBLISHED)

    # Definition of the geometry refinement: the parameter order is
    # the same as the param_names
    calibrant = get_calibrant(multicalib)
    detector = get_detector(multicalib)

    distance = 0.258705917299
    poni1_scale = 0.001
    poni1_offset = 0.132825374721
    poni2_scale = 0.0012272727272727272
    poni2_offset = -0.9488181818181818
    rot1 = 0.00388272369359
    rot2 = -0.00942588451226
    rot3 = 7.19961198098e-07

    parameters = [ Parameter("dist", distance, (distance, distance)),
                   Parameter("poni1_offset", poni1_offset, (0, 0.2)),
                   Parameter("poni1_scale", poni1_scale, (0, 0.002)),
                   Parameter("poni2_offset", poni2_offset, (-1, -0.7)),
                   Parameter("poni2_scale", poni2_scale, (-1, 1)),
                   Parameter("rot1", rot1, (rot1, rot1)),
                   Parameter("rot2", rot2, (rot2, rot2)),
                   Parameter("rot3", rot3, (rot3, rot3))
    ]

    params = {p.name: p.value for p in parameters}
    bounds = {p.name: p.bounds for p in parameters}
    param_names = [p.name for p in parameters]

    # Let's refine poni1 and poni2 also as function of the distance:

    trans_function = GeometryTransformation(param_names=param_names,
                                            pos_names = ["tx", "tz"],
                                            dist_expr="dist",
                                            poni1_expr="tz * poni1_scale + poni1_offset",
                                            poni2_expr="tx * poni2_scale + poni2_offset",
                                            rot1_expr="rot1",
                                            rot2_expr="rot2",
                                            rot3_expr="rot3")

    def pos_function(metadata: MetaData) -> Tuple[float, float]:
        """Definition of the function reading the detector position from the
        header of the image."""
        return metadata.tx, metadata.tz

    gonioref = GoniometerRefinement(params,  # initial guess
                                    bounds=bounds,
                                    pos_function=pos_function,
                                    trans_function=trans_function,
                                    detector=detector,
                                    wavelength=wavelength)

    print("Empty refinement object:")
    print(gonioref)


    # In[20]:

    # Let's populate the goniometer refinement object with the know poni

    with h5py.File(multicalib.filename, mode='r') as h5file:
        for idx in multicalib.idxs:
            base = os.path.splitext(os.path.basename(multicalib.filename))[0]

            label = base + "_%d" % (idx,)
            metadata = get_metadata(h5file, multicalib, idx)
            control_points = os.path.join(PUBLISHED, base + "_%d.npt" % (idx,))
            ai = pyFAI.load(os.path.join(PUBLISHED, base + "_%d.poni" % (idx,)))
            print(ai)

            gonioref.new_geometry(label, metadata.img, metadata, control_points, calibrant, ai)

        print("Filled refinement object:")
        print(gonioref)
        print(os.linesep + "\tlabel \t tx")
        for k, v in gonioref.single_geometries.items():
            print(k, v.get_position())

        for g in gonioref.single_geometries.values():
            ai = gonioref.get_ai(g.get_position())
            print(ai)

        for sg in gonioref.single_geometries.values():
            jupyter.display(sg=sg)

        gonioref.refine2()

    for multi in [multicalib, multicalib2]:
        with h5py.File(multi.filename, mode='r') as h5file:
            optimize_with_new_images(h5file, multi, gonioref, calibrant)

    # for idx, sg in enumerate(gonioref.single_geometries.values()):
    #    sg.geometry_refinement.set_param(gonioref.get_ai(sg.get_position()).param)
    #    jupyter.display(sg=sg)

    gonioref.save(json)

    # pylab.show()


def integrate(json: str) -> None:
    """Integrate a file with a json calibration file"""
    filename = os.path.join(ROOT, "scan_77_01.nxs")
    gonio = pyFAI.goniometer.Goniometer.sload(json)
    wavelength = 4.85945727522e-11
    multicalib = MultiCalib(os.path.join(ROOT, "scan_4_01.nxs"),
                            MetaDataSource("",
                                           H5PathContains("scan_data/actuator_1_1"),
                                           H5PathOptionalItemValue("MARS/D03-1-CX0__DT__DTC_2D-MT_Tz__#1/raw_value", -1.0)),
                            [], "LaB6", "xpad_flat", wavelength)

    with h5py.File(filename, mode='r') as h5file:
        images = []
        positions = []
        for metadata in gen_metadata(h5file, multicalib):
            images.append(metadata.img)
            positions.append((metadata.tx, metadata.tz))
        mai = gonio.get_mg(positions)
        res = mai.integrate1d(images, 10000)
        jupyter.plot1d(res)
        pylab.show()


if __name__ == "__main__":
    JSON = os.path.join(PUBLISHED, "calibration.json")
    # calibration(JSON)

    integrate(JSON)



#!/usr/bin/env python3
""" coding: utf-8
"""

from typing import Iterator, List, NamedTuple, Text, Tuple, Union

import os

from functools import partial

import h5py

from numpy import ndarray

# H5Path data constructors
H5PathContains = NamedTuple("H5PathContains", [("path", Text)])

H5PathOptionalItemValue = NamedTuple('H5OptionalItemValue', [('path', Text),
                                                             ('default', float)])  # noqa
H5PathWithAttribute = NamedTuple("H5PathWithAttribute", [('attribute', Text),
                                                         ('value', bytes)])

H5Path = Union[H5PathContains, H5PathOptionalItemValue, H5PathWithAttribute]

Parameter = NamedTuple("Parameter", [("name", Text),
                                     ("value", float),
                                     ("bounds", Tuple[float, float])])


def _v_attrs(attribute: Text, value: Text, _name: Text, obj) -> h5py.Dataset:
    """extract all the images and accumulate them in the acc variable"""
    if isinstance(obj, h5py.Dataset):
        if attribute in obj.attrs and obj.attrs[attribute] == value:
            return obj


def _v_item(key: Text, name: Text, obj: h5py.Dataset) -> h5py.Dataset:
    if key in name:
        return obj


def get_shape(h5file: h5py.File,
              item: H5Path) -> Tuple:
    res = None
    if isinstance(item, H5PathContains):
        res = h5file.visititems(partial(_v_item, item.path)).shape
    elif isinstance(item, H5PathOptionalItemValue):
        _item = h5file.visititems(partial(_v_item, item.path))
        res = _item.shape if _item else (1,)
    elif isinstance(item, H5PathWithAttribute):
        res = h5file.visititems(partial(_v_attrs,
                                        item.attribute, item.value)).shape
    return res


def get_item_at_index(h5file: h5py.File,
                      item: H5Path,
                      index: int) -> Union[float, ndarray]:
    res = None
    if isinstance(item, H5PathContains):
        res = h5file.visititems(partial(_v_item, item.path))[index]
    elif isinstance(item, H5PathOptionalItemValue):
        _item = h5file.visititems(partial(_v_item, item.path))
        res = _item.value if _item else item.default
    elif isinstance(item, H5PathWithAttribute):
        res = h5file.visititems(partial(_v_attrs,
                                        item.attribute, item.value))[index]
    return res



#!/usr/bin/env python3

import math
import numpy
import matplotlib
matplotlib.use('Agg')

import matplotlib.pyplot as plt

from matplotlib.backends.backend_pdf import PdfPages
from matplotlib import rcParams
from gi.repository import GLib
import gi
gi.require_version("Hkl", "5.0")
from gi.repository import Hkl


def compute_hkl_trajectories(engines, engine, hkl1=None, hkl2=None, n=100):
    """
    compute all the trajectories for a given engine already configured
    """
    if not hkl1:
        hkl1 = [0, 0, 1]
    if not hkl2:
        hkl2 = [0, 1, 1]

    h = numpy.linspace(hkl1[0], hkl2[0], n + 1)
    k = numpy.linspace(hkl1[1], hkl2[1], n + 1)
    l = numpy.linspace(hkl1[2], hkl2[2], n + 1)

    # set the hkl engine and get the results
    trajectories = []
    for hh, kk, ll in zip(h, k, l):
        try:
            solutions = engine.pseudo_axis_values_set([hh, kk, ll],
                                                      Hkl.UnitEnum.USER)
            first_solution = solutions.items()[0]
            for i, item in enumerate(solutions.items()):
                try:
                    trajectories[i]
                except IndexError:
                    trajectories.append([])
                values = item.geometry_get().axis_values_get(Hkl.UnitEnum.USER)
                trajectories[i].append(values)
            engines.select_solution(first_solution)
        except GLib.GError as err:
            pass

    return trajectories


def _plot_legend(axes):
    plt.subplot(3, 4, 1)
    plt.title("legend")
    print("legende", 1)
    for name in axes:
        plt.plot([0, 0], label=name)
    plt.legend()


def plot_hkl_trajectory(filename, geometry, engines,
                        hkl1=None, hkl2=None, n=100):
    """
    plot the trajectory for a engine. It is possible to limit the
    number of trajectory using the max_traj keyword
    """
    axis_names = geometry.axis_names_get()

    hkl = engines.engine_get_by_name("hkl")
    page = 1
    plt.clf()
    plt.suptitle("\"" + filename + "\" " + repr(
        hkl1) + " -> " + repr(hkl2) + " page " + str(page))
    _plot_legend(axis_names)
    idx = 2
    for mode in hkl.modes_names_get():
        hkl.current_mode_set(mode)
        trajectories = compute_hkl_trajectories(engines, hkl, hkl1=hkl1, hkl2=hkl2, n=n)
        print("\"" + filename + "\"", idx, mode, len(trajectories))

        plt.subplot(3, 4, idx)
        plt.title(mode)
        if not len(trajectories):
            plt.text(0.5, 0.5, "Failed", size=20, rotation=0.,
                     ha="center", va="center",
                     bbox=dict(boxstyle="round",
                               ec=(1., 0.5, 0.5),
                               fc=(1., 0.8, 0.8),
                               )
                     )
            plt.draw()
        else:
            plt.ylim(-180, 180)
            if len(trajectories[0]) == 1:
                plt.plot(trajectories[0], 'o-')
            else:
                plt.plot(trajectories[0], '-')

        idx += 1
        if idx > 12:
            pp.savefig()
            plt.clf()
            page += 1
            _plot_legend(axis_names)
            plt.suptitle(filename + " " + repr(
                hkl1) + " -> " + repr(hkl2) + " page " + str(page))
            idx = 2
    pp.savefig()


pp = PdfPages('trajectories.pdf')
rcParams['font.size'] = 6


def main():
    sample = Hkl.Sample.new("toto")
    lattice = Hkl.Lattice.new(1.54, 1.54, 1.54,
                              math.radians(90.),
                              math.radians(90.),
                              math.radians(90.))
    sample.lattice_set(lattice)

    detector = Hkl.Detector.factory_new(Hkl.DetectorType(0))

    for key, factory in Hkl.factories().items():
        geometry = factory.create_new_geometry()
        engines = factory.create_new_engine_list()

        # here we set the detector arm with only positiv values for
        # now tth or delta arm
        for axis in geometry.axis_names_get():
            if axis in ["tth", "delta"]:
                tmp = geometry.axis_get(axis)
                tmp.min_max_set(0, 180., Hkl.UnitEnum.USER)
                geometry.axis_set(axis, tmp)

        engines.init(geometry, detector, sample)

        engines_names = [engine.name_get() for engine in engines.engines_get()]
        if 'hkl' in engines_names:
            plot_hkl_trajectory(key, geometry, engines,
                                hkl1=[0, 0, 1], hkl2=[0, 1, 1], n=100)
    pp.close()

if __name__ == '__main__':
    main()



import pylab
import numpy

FILE = "pinhole1.smv"

with open(FILE, "rb") as f:
    f.seek(512)
    img = numpy.fromfile(f, dtype='uint16', sep="")
    print(img.shape)
    img.shape = ((4096, 4096))
    pylab.imshow(img, interpolate='nearest')
    pylab.show()
